package ru.landgrafhomyak.serdha.api.v0.module

import ru.landgrafhomyak.serdha.api.v0.table.Table
import ru.landgrafhomyak.serdha.api.v0.queries.CopyQuery
import ru.landgrafhomyak.serdha.api.v0.queries.DeleteQuery
import ru.landgrafhomyak.serdha.api.v0.queries.InsertQuery
import ru.landgrafhomyak.serdha.api.v0.queries.SelectQuery
import ru.landgrafhomyak.serdha.api.v0.queries.UpdateQuery
import ru.landgrafhomyak.serdha.api.v0.runtime.SynchronizedDatabase
import ru.landgrafhomyak.serdha.api.v0.runtime.Transaction
import ru.landgrafhomyak.serdha.api.v0.table.CreateTableScope

/**
 * Scope methods for creating new [module template][ModuleTemplate].
 *
 * @see ModuleTemplate.CreateModuleTemplatesScope.createTemplate
 * @see ModuleTemplate.ModuleCreator.createSchema
 */
public interface CreateModuleScope {
	/**
	 * Functional interface for creating table.
	 *
	 * Can be implemented like a companion object that just calls [TableUserExtension]'s constructor with the same or similar signature.
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 *
	 * @see CreateModuleScope.createTable
	 * @see CreateModuleScope.createSessionScopeTemporaryTable
	 * @see CreateModuleScope.createTransactionScopeTemporaryTable
	 * @see UpgradeModuleScope.TableUpgrade
	 */
	public interface TableCreator<TableUserExtension : Any> {
		/**
		 * Scope method that uses [creator] to initialize table schema.
		 *
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with descriptors for future access.
		 *
		 * @see CreateModuleScope.createTable
		 * @see CreateModuleScope.createSessionScopeTemporaryTable
		 * @see CreateModuleScope.createTransactionScopeTemporaryTable
		 * @see UpgradeModuleScope.TableUpgrade
		 */
		public fun createTable(creator: CreateTableScope<TableUserExtension>): TableUserExtension
	}

	/**
	 * Creates table in [specified namespace][namespace] with [specified name][name] and stores it in the database.
	 * If table was created by previous synchronizations, asserts that table at the specified path in the database is same as table provided by [initializer].
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] this table must be [kept][UpgradeModuleScope.upgradeTable] (or [renamed][UpgradeModuleScope.renameTable]) or [deleted][UpgradeModuleScope.deleteTable].
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 * @param namespace Way to group tables if there is a lot.
	 * @param name Name of table for debugging.
	 * @param initializer Table-related descriptors initializer.
	 * @return Table descriptor.
	 */
	public fun <TableUserExtension : Any> createTable(namespace: Namespace, name: String, initializer: TableCreator<TableUserExtension>): Table<TableUserExtension, Nothing>

	/**
	 * Creates a temporary table that exits only until connection to the database not [closed][SynchronizedDatabase._close].
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 * @param namespace Way to group tables if there is a lot.
	 * @param name Name of table for debugging.
	 * @param initializer Table-related descriptors initializer.
	 * @return Table descriptor.
	 */
	public fun <TableUserExtension : Any> createSessionScopeTemporaryTable(namespace: Namespace, name: String, initializer: TableCreator<TableUserExtension>): Table<TableUserExtension, Nothing>

	/**
	 * Creates a temporary table that exits only inside [transaction][Transaction] and auto-deleted (or cleared, depends on driver implementation) when it [finished][Transaction._assertTransactionFinishedAndReleaseResources].
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 * @param namespace Way to group tables if there is a lot.
	 * @param name Name of table for debugging.
	 * @param initializer Table-related descriptors initializer.
	 * @return Table descriptor.
	 */
	public fun <TableUserExtension : Any> createTransactionScopeTemporaryTable(namespace: Namespace, name: String, initializer: TableCreator<TableUserExtension>): Table<TableUserExtension, Nothing>


	/**
	 * Creates 'SELECT' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param QueryUserExtension User's type for containing query-related descriptors.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createTemporarySelectQuery
	 */
	public fun <QueryUserExtension : Any> createSelectQuery(initializer: SelectQuery.Creator<QueryUserExtension>): SelectQuery<QueryUserExtension>

	/**
	 * Creates 'SELECT' query accessible only while the module is [creating][ModuleTemplate.ModuleCreator] or [updating][ModuleTemplate.ModuleUpgrade].
	 * Using this function instead of [CreateModuleScope.createSelectQuery] may help the driver to optimize caches and runtime indexes.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param QueryUserExtension User's type for containing query-related descriptors.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createSelectQuery
	 */
	public fun <QueryUserExtension : Any> createTemporarySelectQuery(initializer: SelectQuery.Creator<QueryUserExtension>): SelectQuery<QueryUserExtension>


	/**
	 * Creates 'INSERT ... VALUES' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TargetTableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param QueryUserExtension User's type for containing query-related descriptors.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createTemporaryInsertQuery
	 */
	public fun <QueryUserExtension : Any, TargetTableUserExtension : Any> createInsertQuery(table: Table<TargetTableUserExtension, *>, initializer: InsertQuery.Creator<QueryUserExtension, TargetTableUserExtension>): InsertQuery<QueryUserExtension>

	/**
	 * Creates 'INSERT ... VALUES' query accessible only while the module is [creating][ModuleTemplate.ModuleCreator] or [updating][ModuleTemplate.ModuleUpgrade].
	 * Using this function instead of [CreateModuleScope.createInsertQuery] may help the driver to optimize caches and runtime indexes.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TargetTableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param QueryUserExtension User's type for containing query-related descriptors.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createInsertQuery
	 */
	public fun <QueryUserExtension : Any, TargetTableUserExtension : Any> createTemporaryInsertQuery(table: Table<TargetTableUserExtension, *>, initializer: InsertQuery.Creator<QueryUserExtension, TargetTableUserExtension>): InsertQuery<QueryUserExtension>

	/**
	 * Creates 'INSERT ... FROM' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TargetTableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createTemporaryCopyQuery
	 */
	public fun <QueryUserExtension : Any, TargetTableUserExtension : Any> createCopyQuery(table: Table<TargetTableUserExtension, *>, initializer: CopyQuery.Creator<QueryUserExtension, TargetTableUserExtension>): CopyQuery<QueryUserExtension>

	/**
	 * Creates 'INSERT ... FROM' query accessible only while the module is [creating][ModuleTemplate.ModuleCreator] or [updating][ModuleTemplate.ModuleUpgrade].
	 * Using this function instead of [CreateModuleScope.createCopyQuery] may help the driver to optimize caches and runtime indexes.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TargetTableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createCopyQuery
	 */
	public fun <QueryUserExtension : Any, TargetTableUserExtension : Any> createTemporaryCopyQuery(table: Table<TargetTableUserExtension, *>, initializer: CopyQuery.Creator<QueryUserExtension, TargetTableUserExtension>): CopyQuery<QueryUserExtension>

	/**
	 * Creates 'UPDATE' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TargetTableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createTemporaryUpdateQuery
	 */
	public fun <QueryUserExtension : Any, TargetTableUserExtension : Any> createUpdateQuery(table: Table<TargetTableUserExtension, *>, initializer: UpdateQuery.Creator<QueryUserExtension, TargetTableUserExtension>): UpdateQuery<QueryUserExtension>

	/**
	 * Creates 'UPDATE' query accessible only while the module is [creating][ModuleTemplate.ModuleCreator] or [updating][ModuleTemplate.ModuleUpgrade].
	 * Using this function instead of [CreateModuleScope.createUpdateQuery] may help the driver to optimize caches and runtime indexes.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TargetTableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createUpdateQuery
	 */
	public fun <QueryUserExtension : Any, TargetTableUserExtension : Any> createTemporaryUpdateQuery(table: Table<TargetTableUserExtension, *>, initializer: UpdateQuery.Creator<QueryUserExtension, TargetTableUserExtension>): UpdateQuery<QueryUserExtension>


	/**
	 * Creates 'DELETE' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TargetTableUserExtension User's type with descriptors related to table from which data will be deleted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createTemporaryDeleteQuery
	 */
	public fun <QueryUserExtension : Any, TargetTableUserExtension : Any> createDeleteQuery(table: Table<TargetTableUserExtension, *>, initializer: DeleteQuery.Creator<QueryUserExtension, TargetTableUserExtension>): DeleteQuery<QueryUserExtension>

	/**
	 * Creates 'DELETE' query accessible only while the module is [creating][ModuleTemplate.ModuleCreator] or [updating][ModuleTemplate.ModuleUpgrade].
	 * Using this function instead of [CreateModuleScope.createDeleteQuery] may help the driver to optimize caches and runtime indexes.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TargetTableUserExtension User's type with descriptors related to table from which data will be deleted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createDeleteQuery
	 */
	public fun <QueryUserExtension : Any, TargetTableUserExtension : Any> createTemporaryDeleteQuery(table: Table<TargetTableUserExtension, *>, initializer: DeleteQuery.Creator<QueryUserExtension, TargetTableUserExtension>): DeleteQuery<QueryUserExtension>

	/**
	 * Replaces [specified namespace][rootNs] with schema provided by [template].
	 * [This namespace][rootNs] will be root namespace for module and shouldn't contain any other definitions.
	 *
	 * @param rootNs Namespace where the new module will be located.
	 * @param template Schema template.
	 * @return Module descriptor.
	 */
	public fun <ModuleUserExtension : Any> substituteModule(rootNs: Namespace, template: ModuleTemplate<ModuleUserExtension>): Module<ModuleUserExtension>
}